/* --------------------------------------------------------- */
/*	BAS Simulator											 */
/* ========================================================= */
/*	(C)2005 t.hara											 */
/* --------------------------------------------------------- */

#include <stdio.h>
#include <windows.h>
#include "bassim.h"

#define	STACK_SIZE	4096

typedef struct {
	unsigned char	*pMemory;
	unsigned int	Size;
	unsigned int	InstPtr;
	unsigned int	StackPtr;
	int				FuncRes;
	int				Stack[ STACK_SIZE ];
} BASSIM_T;

enum {
	CMD_pusl = 0, CMD_pusv, CMD_pusc, CMD_pres,
	CMD_apus, CMD_pop_, CMD_apop, CMD_lnot,
	CMD_not_, CMD_mul_, CMD_div_, CMD_mod_,
	CMD_add_, CMD_sub_, CMD_shr_, CMD_shl_,
	CMD_and_, CMD_or__, CMD_xor_, CMD_equ_,
	CMD_gequ, CMD_lequ, CMD_gret, CMD_less,
	CMD_nequ, CMD_land, CMD_lor_, CMD_call,
	CMD_ret_, CMD_sysc, CMD_jump, CMD_jmpz,
	CMD_jmpt, CMD_halt, CMD_MAX,
};

#define	GET_WORD( p )		GetWord( pInstance, (p) )
#define	SET_WORD( p, a )	SetWord( pInstance, (p), (a) )
#define PUSH( a )			Push( pInstance, (a) )
#define POP()				Pop( pInstance )

typedef int (*SYSTEM_CALL_T)( unsigned char *pMemory, int *pParams );

#ifdef _DEBUG
	static const char *pCommand[256] = {
		"pusl", "pusv", "pusc", "pres", 
		"apus", "pop_", "apop", "lnot", 
		"not_", "mul_", "div_", "mod_", 
		"add_", "sub_", "shr_", "shl_", 
		"and_", "or__", "xor_", "equ_", 
		"gequ", "lequ", "gret", "less", 
		"nequ", "land", "lor_", "call", 
		"ret_", "sysc", "jump", "jmpz", 
		"jmpt", "halt",  
	};
#endif

/* --------------------------------------------------------- */
static int GetWord( BASSIM_T *pInstance, unsigned int Index ) {

	if( pInstance->Size <= Index ) {
		printf( "READ 0x%08X\n", Index );
		return 0;
	}
	return( *(int*)(pInstance->pMemory + Index) );
}

/* --------------------------------------------------------- */
static void SetWord( BASSIM_T *pInstance, unsigned int Index, int Data ) {

	if( pInstance->Size <= Index ) {
		printf( "WRITE 0x%08X <== 0x%08X\n", Index, Data );
		return;
	}
	*(int*)(pInstance->pMemory + Index) = Data;
}

/* --------------------------------------------------------- */
static void Push( BASSIM_T *pInstance, int Data ) {

	if( pInstance->StackPtr >= STACK_SIZE ) {
		printf( "STACK OVERFLOW!!\n" );
		return;
	}
	pInstance->Stack[ pInstance->StackPtr ] = Data;
	pInstance->StackPtr ++;
}

/* --------------------------------------------------------- */
static int Pop( BASSIM_T *pInstance ) {

	if( pInstance->StackPtr <= 0 ) {
		printf( "STACK UNDERFLOW!!\n" );
		return 0;
	}
	pInstance->StackPtr --;
	return pInstance->Stack[ pInstance->StackPtr ];
}

/* --------------------------------------------------------- */
static int _Print( unsigned char *pMemory, int *pParam ) {

	printf( "%s", pMemory + pParam[0] );
	return 0;
}

/* --------------------------------------------------------- */
static int _PrintDec( unsigned char *pMemory, int *pParam ) {

	printf( "%d", pParam[0] );
	return 0;
}

/* --------------------------------------------------------- */
static int _PrintHex( unsigned char *pMemory, int *pParam ) {

	printf( "%04X", pParam[0] );
	return 0;
}

/* --------------------------------------------------------- */
static int _Abs( unsigned char *pMemory, int *pParam ) {

	if( pParam[0] < 0 ) return -pParam[0];
	return pParam[0];
}

/* --------------------------------------------------------- */
static int _Rnd( unsigned char *pMemory, int *pParam ) {

	return (int) rand();
}

/* --------------------------------------------------------- */
static int _Sleep( unsigned char *pMemory, int *pParam ) {

	Sleep( pParam[0] );
	return 0;
}

/* --------------------------------------------------------- */
static int _SetIntSub( unsigned char *pMemory, int *pParam ) {

	/* H */
	return 0;
}

/* --------------------------------------------------------- */
static const SYSTEM_CALL_T SystemCallTable[] = {
	_Print,
	_PrintDec,
	_PrintHex,
	_Abs,
	_Rnd,
	_Sleep,
	_SetIntSub,
};

/* --------------------------------------------------------- */
static void _SystemCall( BASSIM_T *pInstance, int Params, int Function ) {
	int Param[ 256 ];
	int i;
	unsigned char *pMemory;
	int *pStack;

	pMemory	= pInstance->pMemory;
	pStack	= pInstance->Stack;
	for( i = 0; i < Params; i++ ) {
		Param[i] = POP();
	}
	if( (unsigned int)Function > sizeof(SystemCallTable)/sizeof(SystemCallTable[0]) ) {
		fprintf( stderr, "Invalid system call.\n" );
		return;
	}
	pInstance->FuncRes = (SystemCallTable[ Function ])( pMemory, Param );
}

/* --------------------------------------------------------- */
static void _Simulator( BASSIM_T *pInstance ) {
	unsigned char Cmd;
	unsigned char *pMemory;
	int *pStack, Param1, Param2;

	pMemory	= pInstance->pMemory;
	pStack	= pInstance->Stack;
	for( ;; ) {
		if( pInstance->InstPtr >= pInstance->Size ) {
			printf( "INVALID CODE MEMORY 0x%08X\n", pInstance->InstPtr );
			return;
		}
		Cmd = pMemory[ pInstance->InstPtr ];
#		ifdef _DEBUG
			printf( "%08X %s\n", pInstance->InstPtr, pCommand[ Cmd ] );
#		endif
		switch( Cmd ) {
		case CMD_pusl:
			PUSH( GET_WORD( pInstance->InstPtr + 1 ) );
			pInstance->InstPtr += 5;
			break;
		case CMD_pusv:
			PUSH( GET_WORD( GET_WORD( pInstance->InstPtr + 1 ) ) );
			pInstance->InstPtr += 5;
			break;
		case CMD_pusc:
			PUSH( GET_WORD( pInstance->InstPtr + 1 ) );
			pInstance->InstPtr += 5;
			break;
		case CMD_pres:
			PUSH( pInstance->FuncRes );
			pInstance->InstPtr++;
			break;
		case CMD_apus:
			Param1 = GET_WORD( GET_WORD( pInstance->InstPtr + 1 ) + POP() * 4 );
			PUSH( Param1 );
			pInstance->InstPtr += 5;
			break;
		case CMD_pop_:
			SET_WORD( GET_WORD( pInstance->InstPtr + 1 ), POP() );
			pInstance->InstPtr += 5;
			break;
		case CMD_apop:
			Param1 = GET_WORD( pInstance->InstPtr + 1 ) + POP();
			SET_WORD( Param1, POP() * 4 );
			pInstance->InstPtr += 5;
			break;
		case CMD_lnot:
			Param1 = !POP();
			PUSH( Param1 );
			pInstance->InstPtr++;
			break;
		case CMD_not_:
			Param1 = ~POP();
			PUSH( Param1 );
			pInstance->InstPtr++;
			break;
		case CMD_mul_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 * Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_div_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 / Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_mod_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 % Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_add_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 + Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_sub_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 - Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_shr_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 >> Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_shl_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 << Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_and_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 & Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_or__:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 | Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_xor_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 ^ Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_equ_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 == Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_gequ:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 >= Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_lequ:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 <= Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_gret:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 > Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_less:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 < Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_nequ:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 != Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_land:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 && Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_lor_:
			Param2 = POP();
			Param1 = POP();
			PUSH( Param1 || Param2 );
			pInstance->InstPtr++;
			break;
		case CMD_call:
			PUSH( pInstance->InstPtr + 5 );
			pInstance->InstPtr = GET_WORD( pInstance->InstPtr + 1 );
			break;
		case CMD_ret_:
			pInstance->InstPtr = POP();
			break;
		case CMD_sysc:
			Param1 = pMemory[ pInstance->InstPtr + 1 ];
			Param2 = GET_WORD( pInstance->InstPtr + 2 );
			_SystemCall( pInstance, Param1, Param2 );
			pInstance->InstPtr += 6;
			break;
		case CMD_jump:
			pInstance->InstPtr = GET_WORD( pInstance->InstPtr + 1 );
			break;
		case CMD_jmpz:
			if( POP() == 0 ) {
				pInstance->InstPtr = GET_WORD( pInstance->InstPtr + 1 );
			}
			else {
				pInstance->InstPtr += 5;
			}
			break;
		case CMD_jmpt:
			if( POP() != 0 ) {
				pInstance->InstPtr = GET_WORD( pInstance->InstPtr + 1 );
			}
			else {
				pInstance->InstPtr += 5;
			}
			break;
		case CMD_halt:
			return;
		default:
			fprintf( stderr, "Invalid operation code 0x%02X at %08X.\n", Cmd, pInstance->InstPtr );
			return;
		}
	}
}

/* --------------------------------------------------------- */
void bassimRun( unsigned char *pMemory, int Size ) {
	BASSIM_T Bassim;

	Bassim.pMemory	= pMemory;
	Bassim.InstPtr	= 0;
	Bassim.StackPtr	= 0;
	Bassim.FuncRes	= 0xAA55;
	Bassim.Size		= Size;

	_Simulator( &Bassim );
}
